/* $Id: raw2mq.c,v 1.3 2023/02/23 06:17:41 urabe Exp $ */
/* "raw2mq.c"   2023.2.5 urabe */
/*              modified from raw_ch.c */
/*              2023.2.5-2.19 raw2mq.c */
/*              2023.2.23 daemon mode */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <mosquitto.h>

#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else  /* !TIME_WITH_SYS_TIME */
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else  /* !HAVE_SYS_TIME_H */
#include <time.h>
#endif  /* !HAVE_SYS_TIME_H */
#endif  /* !TIME_WITH_SYS_TIME */

#include "winlib.h"
#include "daemon_mode.h"
#define PORT 1883
#define SPORT 8883
#define CAFILE "/etc/ssl/cert.pem"
#define CAPATH "/etc/ssl/certs"

static const char rcsid[] =
  "$Id: raw2mq.c,v 1.3 2023/02/23 06:17:41 urabe Exp $";

char *progname,*logfile;
static int daemon_mode;
int syslog_mode,exit_status;
int debug=0;
int connected=0;
char tb[200];
struct mosquitto *mosq;
time_t ltime_p;
unsigned long n_packets,n_bytes;   /* 64bit ok */

void
report()
{
  time_t ltime,j_timt,k_timt;

  time(&ltime);
  j_timt=ltime-ltime_p;
  if(j_timt != 0) {
    k_timt=j_timt/2;
    if(ltime_p)
      {
      snprintf(tb,sizeof(tb),
	"%lu msgs %lu B in %ld s ( %ld msgs/s %ld B/s ) ",
	 n_packets,n_bytes,(long)j_timt,
	 (n_packets+k_timt)/j_timt,(n_bytes+k_timt)/j_timt);
      write_log(tb);
      n_packets=n_bytes=0;
      }
    ltime_p=ltime;
  }
}

void
endprog()
{
  mosquitto_destroy(mosq);
  mosquitto_lib_cleanup();
  end_program();
}

void
usage(void)
{
  WIN_version();
  fprintf(stderr, "%s\n", rcsid);
  if(daemon_mode)
    fprintf(stderr,
  " usage : '%s (-gs) (-t [topic]) (-p [port]) (-q [qos]) (-i [id])\\\n\
               (-c [cafile] / -d [capath]) (-C [certfile] -k [keyfile])\\\n\
               (-u [user]  -P [pswd])\\\n\
             [in_key] [host] ([log file])'\n",progname);
  else
    fprintf(stderr,
  " usage : '%s (-gsD) (-t [topic]) (-p [port]) (-q [qos]) (-i [id])\\\n\
               (-c [cafile] / -d [capath]) (-C [certfile] -k [keyfile])\\\n\
               (-u [user]  -P [pswd])\\\n\
             [in_key] [host] ([log file])'\n",progname);
}

void
on_connect(struct mosquitto *mosq,void *obj,int rc)
{
  if(rc) {
    snprintf(tb,sizeof(tb),"%s",mosquitto_connack_string(rc));
    write_log(tb);
    endprog();
  }
  connected=1;
  write_log("connected");
}

void
on_disconnect(struct mosquitto *mosq,void *obj,int rc)
{
  connected=0;
  write_log("disconnected");
}

int
main(int argc, char *argv[])
  {
  struct Shm *shr;
  key_t rawkey;
  uint32_w uni;
  uint8_w *ptr,*ptr_lim,*ptr_save;
  int c,i,tow,ret,rc;
  int eobsize_in, eobsize_in_count;
  uint32_w size;
  unsigned long c_save;
  uint16_t port=PORT;
  char *host=NULL,*id=NULL,*user=NULL,*pswd=NULL,*capath=NULL,
    *certfile=NULL,*keyfile=NULL,*cafile=NULL,*topic="win/data";
  int qos=0;
  int ssl=0;
  unsigned int mid=0;
  int portset=0;
  int ca_default=1;
  n_packets=n_bytes=0;

  if((progname=strrchr(argv[0],'/')) != NULL) progname++;
  else progname=argv[0];

  daemon_mode=syslog_mode=0;
  exit_status=EXIT_SUCCESS;
  if(strcmp(progname,"raw2mqd")==0) daemon_mode=1;

  while((c=getopt(argc,argv,"c:C:d:Dgi:k:p:P:q:st:u:"))!=-1)
    {
    switch(c)
      {
      case 'c':   /* cafile */
        cafile=optarg;
        ca_default=0;
        break;
      case 'C':   /* certfile */
        certfile=optarg;
        break;
      case 'd':   /* capath */
        capath=optarg;
        ca_default=0;
        break;
      case 'D':   /* daemon mode */
        daemon_mode = 1;
        break;
      case 'g':   /* debug print */
        debug++;
        break;
      case 'i':   /* unique client id */
        id=optarg;
        break;
      case 'k':   /* keyfile */
        keyfile=optarg;
        break;
      case 'p':   /* port */
        port=(uint16_t)atoi(optarg);
        portset=1;
        break;
      case 'P':   /* password */
        pswd=optarg;
        break;
      case 'q':   /* qos */
        qos=atoi(optarg);
        break;
      case 's':   /* use ssl/tls */
        ssl=1;
        break;
      case 't':   /* topic */
        topic=optarg;
        break;
      case 'u':   /* username */
        user=optarg;
        break;
      default:
        fprintf(stderr," option -%c unknown\n",c);
        usage();
        exit(1);
      }
    }
  optind--;
  if(argc<3+optind)
    {
    usage();
    exit(1);
    }
  rawkey=atol(argv[1+optind]);
  host=argv[2+optind];
  if(argc>3+optind) logfile=argv[3+optind];
  else {
    logfile=NULL;
    if(daemon_mode) syslog_mode=1;
  }

  /* daemon mode */
  if (daemon_mode) {
    daemon_init(progname, LOG_USER, syslog_mode);
    umask(022);
  }

  if(qos<0 || qos>2) {
    write_log("invalid QoS value");
    exit(1);
  }

  if(ssl && !portset) port=SPORT;

  write_log("start");

  mosquitto_lib_init();
  if(id==NULL) mosq=mosquitto_new(NULL,1,NULL);
  else mosq=mosquitto_new(id,0,NULL);
  if(!mosq) {
    write_log("failed to create mosquitto");
    endprog();
  }

  mosquitto_username_pw_set(mosq,user,pswd);

  if(ssl) {
    if(ca_default) {
      rc=mosquitto_tls_set(mosq,cafile=CAFILE,capath,certfile,keyfile,NULL);
      if(rc!=MOSQ_ERR_SUCCESS) {
        rc=mosquitto_tls_set(mosq,cafile=NULL,capath=CAPATH,certfile,keyfile,NULL);
        if(rc!=MOSQ_ERR_SUCCESS) {
          snprintf(tb,sizeof(tb),"mosquitto_tls_set failed(%d)\n",rc);
          write_log(tb);
          endprog();
        }
      }
    } else {
      rc=mosquitto_tls_set(mosq,cafile,capath,certfile,keyfile,NULL);
      if(rc!=MOSQ_ERR_SUCCESS) {
        snprintf(tb,sizeof(tb),"mosquitto_tls_set failed(%d)\n",rc);
        write_log(tb);
        endprog();
      }
    }
  }

  if(ssl) snprintf(tb,sizeof(tb),"mqtts://%s:%s@%s:%d topic=%s qos=%d",
    user,pswd,host,port,topic,qos);
  else snprintf(tb,sizeof(tb),"mqtt://%s:%s@%s:%d topic=%s qos=%d",
    user,pswd,host,port,topic,qos);
  write_log(tb);
  if(ssl) {
    if(cafile) {snprintf(tb,sizeof(tb),"cafile=%s",cafile);write_log(tb);}
    if(capath) {snprintf(tb,sizeof(tb),"capath=%s",capath);write_log(tb);}
    if(certfile) {snprintf(tb,sizeof(tb),"certfile=%s",certfile);write_log(tb);}
    if(keyfile) {snprintf(tb,sizeof(tb),"keyfile=%s",keyfile);write_log(tb);}
  }
  if(id) {snprintf(tb,sizeof(tb),"id=%s",id);write_log(tb);}

  mosquitto_connect_callback_set(mosq,on_connect);
  mosquitto_disconnect_callback_set(mosq,on_disconnect);

  if(mosquitto_connect_bind(mosq,host,port,60,NULL)) {
    strerror_r(errno,tb,sizeof(tb));
    write_log("failed to connect broker");
    endprog();
  }
  if(mosquitto_loop_start(mosq)!=MOSQ_ERR_SUCCESS){
    write_log("Unable to start loop");
    endprog();
  }
  sleep(1);

  /* shared memory */
  shr=Shm_read(rawkey,"in");

  signal(SIGTERM,(void *)endprog);
  signal(SIGINT,(void *)endprog);
  signal(SIGHUP,(void *)report);
  time(&ltime_p);

reset:
  while(shr->r==(-1)) sleep(1);
  ptr=shr->d+shr->r;
  tow=(-1);

  size=mkuint4(ptr);
  if(mkuint4(ptr+size-4)==size) eobsize_in=1;
  else eobsize_in=0;
  eobsize_in_count=eobsize_in;
  snprintf(tb,sizeof(tb),"eobsize_in=%d",eobsize_in);
  write_log(tb);

  for(;;) {
    size=mkuint4(ptr_save=ptr);
    if(mkuint4(ptr+size-4)==size) {
      if(++eobsize_in_count==0) eobsize_in_count=1;
    }
    else eobsize_in_count=0;
    if(eobsize_in && eobsize_in_count==0) goto reset;
    if(!eobsize_in && eobsize_in_count>3) goto reset;
    ptr_lim=ptr+size;
    if(eobsize_in) ptr_lim-=4;
    if(debug>1) {
      printf("    ");
      for(i=0;i<16;i++) printf("%02X",ptr[i]);
      printf(" :%6d b read\n",size);
    }

    c_save=shr->c;
    ptr+=4;
  /* make output data */
    uni=(uint32_w)(time(NULL)-TIME_OFFSET);
    i=uni-mkuint4(ptr);
    if(i>=0 && i<1440)   /* with tow */
      {
      if(tow!=1)
        {
	snprintf(tb,sizeof(tb),"with TOW (diff=%ds)",i);
        write_log(tb);
        if(tow==0)
          {
          write_log("reset");
          goto reset;
          }
        tow=1;
        }
      ptr+=4;
      }
    else if(tow!=0)
      {
      write_log("without TOW");
      if(tow==1)
        {
        write_log("reset");
        goto reset;
        }
      tow=0;
      }

    if(connected) {
      ret=mosquitto_publish(mosq,&mid,topic,ptr_lim-ptr,ptr,qos,0);
      if(ret!=MOSQ_ERR_SUCCESS) {
        write_log("failed to publish");
      }
      if(debug) {
        for(i=0;i<16;i++) printf("%02X",ptr[i]);
        printf(" :%6d b sent topic=%s qos=%d mid=%d\n",ptr_lim-ptr,topic,qos,mid);
      }
      mid++;
      n_packets++;
      n_bytes+=ptr_lim-ptr;
    }
    if((ptr=ptr_save+size)>shr->d+shr->pl) ptr=shr->d;
    while(ptr==shr->d+shr->p) usleep(10000);
    if(shr->c<c_save || mkuint4(ptr_save)!=size) {
      write_log("reset");
      goto reset;
    }
  }
}
